import * as THREE from '../build/three.module.js';
import { EffectComposer } from './jsm/postprocessing/EffectComposer.js';
import { RenderPass } from './jsm/postprocessing/RenderPass.js';
import { FilmPass } from './jsm/postprocessing/FilmPass.js';
import { BloomPass } from './jsm/postprocessing/BloomPass.js';

const ENTIRE_SCENE = 0, BLOOM_SCENE = 1;

class App {
    constructor() {
        const domWebGL = document.createElement('div');
        document.body.appendChild(domWebGL);
        this._domWebGL = domWebGL;

        const renderer = new THREE.WebGLRenderer({ antialias: true });
        renderer.outputEncoding = THREE.sRGBEncoding;
        renderer.setPixelRatio(window.devicePixelRatio);
        //renderer.shadowMap.enabled =  true;
        //renderer.toneMapping = THREE.ReinhardToneMapping;
        renderer.autoClear = false;
        domWebGL.appendChild(renderer.domElement);

        this._renderer = renderer;

        const scene = new THREE.Scene();
        this._scene = scene;

        this._clock = new THREE.Clock();


        this._setupCamera();
        //this._setupLight();
        this._setupPostprocessing();
        //this._setupInteractions();
        //this._setupGUI();
        this._setupModel();
        //this._setupControls();
        //this._setupHelper();

        window.onresize = this.resize.bind(this);

        this.resize();
        this.render()
    }

    _setupInteractions() {
    }

    _setupPostprocessing() {
        const renderModel = new RenderPass(this._scene, this._camera);
        const effectBloom = new BloomPass(1.24);
        const effectFilm = new FilmPass(0.35, 0.95, 2048, false);
        const composer = new EffectComposer(this._renderer);

        composer.addPass(renderModel);
        composer.addPass(effectBloom);
        composer.addPass(effectFilm);

        this._composer = composer;
    }

    _setupModel() {

        const vertexShader = /* glsl */ `
			uniform vec2 uvScale;
			varying vec2 vUv;

			void main()
			{
				vUv = uvScale * uv;
				vec4 mvPosition = modelViewMatrix * vec4( position, 1.0 );
				gl_Position = projectionMatrix * mvPosition;
			}
        `;

        const fragmentShader = /* glsl */`
			uniform float time;
            uniform float fogDensity;
            uniform vec3 fogColor;
            uniform sampler2D texture1;
            uniform sampler2D texture2;
            varying vec2 vUv;

            void main( void ) {
                vec2 position = -1.0 + 2.0 * vUv;
                vec4 noise = texture2D( texture1, vUv );
                vec2 T1 = vUv + vec2( 1.5, - 1.5 ) * time * 0.02;
                vec2 T2 = vUv + vec2( - 0.5, 2.0 ) * time * 0.01;

                T1.x += noise.x * 2.0;
                T1.y += noise.y * 2.0;
                T2.x -= noise.y * 0.2;
                T2.y += noise.z * 0.2;

                float p = texture2D( texture1, T1 * 2.0 ).a;

                vec4 color = texture2D( texture2, T2 * 2.0 );
                vec4 temp = color * ( vec4( p, p, p, p ) * 2.0 ) + ( color * color - 0.1 );

                if( temp.r > 1.0 ) { temp.bg += clamp( temp.r - 2.0, 0.0, 100.0 ); }
                if( temp.g > 1.0 ) { temp.rb += temp.g - 1.0; }
                if( temp.b > 1.0 ) { temp.rg += temp.b - 1.0; }

                gl_FragColor = temp;

                float depth = gl_FragCoord.z / gl_FragCoord.w;
                const float LOG2 = 1.442695;
                float fogFactor = exp2( - fogDensity * fogDensity * depth * depth * LOG2 );
                fogFactor = 1.0 - clamp( fogFactor, 0.0, 1.0 );

                gl_FragColor = mix( gl_FragColor, vec4( fogColor, gl_FragColor.w ), fogFactor );
            }
        `;

        const textureLoader = new THREE.TextureLoader();

        const uniforms = {
            "fogDensity": { value: 0.45 },
            "fogColor": { value: new THREE.Vector3(0, 0, 0) },
            "time": { value: 1.0 },
            "uvScale": { value: new THREE.Vector2(3.0, 1.0) },
            "texture1": { value: textureLoader.load("textures/lava/cloud.png") },
            "texture2": { value: textureLoader.load("textures/lava/lavatile.jpg") }
        };

        uniforms["texture1"].value.wrapS = uniforms["texture1"].value.wrapT = THREE.RepeatWrapping;
        uniforms["texture2"].value.wrapS = uniforms["texture2"].value.wrapT = THREE.RepeatWrapping;

        const size = 0.65;

        const material = new THREE.ShaderMaterial( {uniforms, vertexShader, fragmentShader} );
        const mesh = new THREE.Mesh(new THREE.TorusGeometry(size, 0.3, 30, 30), material);
        mesh.rotation.x = 0.3;
        this._scene.add(mesh);

        this._uniforms = uniforms;
        this._mesh = mesh;
    }

    _setupHelper() {
        //.
    }

    _setupGUI() {}

    _updateMaterial() {
        const uniforms = this._mesh.material.uniforms;
        uniforms.threshold.value = this._parameters.threshold;
        uniforms.steps.value = this._parameters.steps;
    }

    _setupCamera() {
        const camera = new THREE.PerspectiveCamera(35, window.innerWidth / window.innerHeight, 1, 3000);
        camera.position.set(0, 0, 4);
        this._camera = camera;
    }

    _setupControls() {
        //.
    }

    _setupLight() {
    }

    update() {
        const delta = 5 * this._clock.getDelta();
        this._uniforms["time"].value += 0.2 * delta;
        this._mesh.rotation.y += 0.0125 * delta;
        this._mesh.rotation.x += 0.05 * delta;
    }

    render() {
        requestAnimationFrame(this.render.bind(this));
        //this._renderer.render(this._scene, this._camera);   
        this._renderer.clear();
        this._composer.render();

        this.update();
    }

    resize() {
        const width = window.innerWidth;
        const height = window.innerHeight;

        this._camera.aspect = width / height;
        this._camera.updateProjectionMatrix();
        
        this._renderer.setSize(width, height);
        this._composer.setSize(width, height);
    }
}

window.onload = function () {
    new App();
}